<?php

/**
 * Tools to synchronize SQL tables (contents and definitions)
 *
 * @author Yvan Taviaud, from original Patrick Poulain's code @ Labsmedia.
 * @see http://petitchevalroux.net/index.html http://www.labsmedia.com/index.html
 * @since 15/12/2008
 */
class Database_Checker
{
	private $connections = array();
	private $aliases = array();
	private $skip = array();
	private $distinctrows = array();
	private $database = array();
	private $server = array();
	private $servers = array();
	private $uid = array();
	private $format = 'sql';
	private $charset = 'UTF8';
	private $extension = 'mysql';

	/** @var integer keys Number of rows to retrieve while checking tables differencies (selection of keys) */
	const keys = 2000;
	/** @var integer rows Number of rows to retrieve while checking tables differencies (selection of data) */
	const rows = 100;
	/** @var integer diffs Number of diff' passes before stopping the search for differencies, when one block of rows has been found with differencies */
	const diffs = 3;

	/** @var integer $memory Maximum memory in MB to use before stopping the search for differencies. Once this amount of memory *MAY* be used, no more table differencies are done */
	private $memory = 10;

	/** @var integer lru Number of tables to keep in RAM (cache for speed) */
	const lru = 100;
	const host = 1;
	const user = 2;
	const pass = 3;
	const http = 4;
	/** @var boolean Set to true to enable debugging (see $this->getInfos() then) */
	const debug = false;
	/** @var string Characters used to join and split the PK (if your PK has many columns) */
	const pkChars = '*!!*';
	/** @var boolean Disable the setting of auto_increment (this adds ALTER TABLE which is damn slow, and INSERT statements can update the value if the AUTO_INCREMENT column is added with a higher value) */
	const autoincrement = false;
	/** @var boolean Set to false to disable checksum of tables, as this may lead the script to think that tables are the same, but in fact these are not (see: http://dev.mysql.com/doc/refman/5.0/en/checksum-table.html) */
	const checksum = true;

	/** @var mixed $rowNumberCacheDir File, where we cache the last row number we look at while checking for table data differencies. Use it to improve the time spent on comparing rows, by avoiding check on just checked rows. The number cached is ignored after 1 hour of no usage (for the next full check) */
	private $rowNumberCacheDir = '';

	/** @var integer Microseconds to wait between 2 HTTP calls, to limit load on the distant server (1,000,000 = 1 second)  */
	const wait = 0;
	/** @var integer HTTP timeout in seconds */
	const timeout = 60;
	/** @var integer inserts Group INSERT queries to speed up INSERT execution, 0 to disable */
	const inserts = 100;

	private $errors = array();
	private $infos = array();

	/** @var string $version Revision number of this class */
	const version = '0.14';

	/** @var array $keys Order of array keys to run the SQL commands in the good order, to use with array_combine() */
	public $keys = array('drop_idx', 'add', 'change', 'drop', 'create', 'add_idx', 'change_idx', 'order', 'delete', 'insert', 'update', 'auto');

	public function __construct()
	{
		if (function_exists('mysqli_connect'))
		{
			$this->extension = 'mysqli';
		}
		elseif (function_exists('mysql_connect'))
		{
			$this->extension = 'mysql';
		}
		else
		{
			trigger_error('No MySQL extension found!', E_USER_ERROR);
		}
	}

	public function setFormat($format)
	{
		if ($format !== 'sql' && $format !== 'html' && $format !== 'array')
		{
			$format = 'sql';
		}
		$this->format = $format;
		if (self::diffs <= 0)
		{
			$this->errors[] = '-- self::diffs should not be null';
		}
	}

	/**
	 * Sets server credentials
	 *
	 * @param string $name		Unique name
	 * @param string $host		Host (ie: localhost)
	 * @param string $user		User
	 * @param string $password	Password
	 * @param string $url		Url (to be used in distant mode)
	 */
	public function addServer($name, $host, $user, $password, $url = false)
	{
		if (isset($this->servers[$name]))
		{
			return false;
		}
		if ($url !== false && strpos($url, '?') === false)
		{
			$url .= '?';
		}
		$this->servers[$name] = array(self::host => $host, self::user => $user, self::pass => $password, self::http => $url);
		return true;
	}

	/**
	 * Sets a database
	 *
	 * @param string $server	Server to use
	 * @param string $database	Database name
	 * @param string $source	Use as source or as destination
	 */
	public function selectDatabase($server, $database, $source)
	{
		if (!isset($this->servers[$server]))
		{
			$this->errors[] = '-- No server «'.$server.'» was defined, can\'t use it in selectDatabase()';
			return false;
		}
		$this->database[$source] = $database;
		$this->server[$source] = $server;
		$this->uid[$source] = '.'.$database.'.'.$server;
		return true;
	}

	/**
	 * Adds a table alias for another table (if database1.table1 is the same as database2.table2 for instance)
	 *
	 * ALWAYS add tables by groups of foreign tables (if there's foreign keys), for example if table A is linked to B, and these tables are A1/B1 and A2/B2 in the other database, add the link A/A1 and B/B1 before the links A/A2 and B/B2
	 *
	 * @param string $table1 Table in the source database
	 * @param string $table2 Table in the destination database
	 */
	public function addTableAlias($table1, $table2, $addReverse = true)
	{
		static $index = array();
		if ($addReverse === true)
		{
			$id = $this->uid[true].$this->uid[false];
		}
		else
		{
			$id = $this->uid[false].$this->uid[true];
		}
		if (!isset($this->aliases[$id][$table1]))
		{
			$this->aliases[$id][$table1] = array();
		}
		if (!isset($index[$id]))
		{
			$index[$id] = 0;
		}
		if (isset($this->aliases[$id][$table1][$index[$id]]))
		{
			$index[$id]++;
		}
		$this->aliases[$id][$table1][$index[$id]] = $table2;
		if ($addReverse === true)
		{
			$this->infos[] = '-- Added alias '.$this->database[false].'.'.$table2.' for '.$this->database[true].'.'.$table1;
			$this->addTableAlias($table2, $table1, false);
		}
		return true;
	}

	/**
	 * Skips columns for table's content differences
	 *
	 * @param string $table Table
	 * @param array $column Column to be skipped
	 */
	public function skipColumnDiff($table, $column)
	{
		$this->skip[$table][$column] = true;
		return true;
	}

	/**
	 * Declare a table without PK nor UNIQUE key to have only distinct rows (useful for table's data comparison)
	 *
	 * @param string $table Table
	 * @param array  $cols  Columns to ignore in the distinct
	 */
	public function setDistinctRows($table, $cols = array())
	{
		$this->distinctrows[$table] = array_flip($cols);
		return true;
	}

	/**
	 * Set the file where to keep the row's number, to speed up comparison of data on large tables.
	 *
	 * @param string $file
	 */
	public function setRowNumberCacheDirectory($directory)
	{
		if (!is_dir($directory))
		{
			$this->errors[] = '-- Directory used for caching row\'s number doesn\'t exist. Please create it first.';
			return false;
		}
		if (strtolower(substr(PHP_OS, 0, 3)) === 'win')
		{
			$this->rowNumberCacheDir = rtrim($directory, '\\').'\\';
		}
		else
		{
			$this->rowNumberCacheDir = rtrim($directory, '/').'/';
		}
	}

	/**
	 * Connect to MySQL (or open an HTTP socket for distant mode)
	 *
	 * @param boolean $source
	 * @return resource
	 */
	private function getConnection($source)
	{
		$server = $this->server[$source];
		if (!isset($this->servers[$server]))
		{
			exit('Error: server '.$server.' not set, use addServer()');
		}
		if (!isset($this->connections[$server]))
		{
			$this->connections[$server] = false;
		}

		if ($this->connections[$server] !== false)
		{
			return $this->connections[$server];
		}

		if (self::debug === true)
		{
			$this->infos[] = '-- Connection to '.($source === true ? 'source' : 'destination').': opening MySQL link on '.$this->servers[$server][self::host];
		}

		/* Forcing the MySQL link to be a new one, so that the mysql_select_db doesn't change the database of the source link */
		if ($this->extension === 'mysql')
		{
			$this->connections[$server] = mysql_connect($this->servers[$server][self::host], $this->servers[$server][self::user], $this->servers[$server][self::pass]);
		}
		else
		{
			$this->connections[$server] = mysqli_connect($this->servers[$server][self::host], $this->servers[$server][self::user], $this->servers[$server][self::pass]);
		}
		if ($this->connections[$server] === false)
		{
			if ($this->extension === 'mysql')
			{
				$this->errors[] = '-- Connection to MySQL server failed: '.mysql_error();
			}
			else
			{
				$this->errors[] = '-- Connection to MySQL server failed: '.$this->connections[$server]->error;
			}
		}
		else
		{
			if ($this->extension === 'mysql')
			{
				mysql_query('SET NAMES '.$this->charset);
			}
			else
			{
				$this->connections[$server]->query('SET NAMES '.$this->charset);
			}
		}
		return $this->connections[$server];
	}

	/**
	 * Run a query against the database, with special keywords linked to specific queries (this is to avoid passing the SQL statement in the GET requests of the distant mode)
	 *
	 * @param boolean $source
	 * @param string $query
	 * @param array $params
	 * @return array
	 */
	private function query($source, $query, $params = array())
	{
		$server = $this->servers[$this->server[$source]];
		if ($server[self::http] !== false)
		{
			if (self::debug === true)
			{
				$this->infos[] = '-- Request on '.($source === true ? 'source' : 'destination').': '.$query.' ('.implode(',', $params).')';
			}
			$http = Http::getInstance();
			$http->usleep = self::wait;
			$http->timeout = self::timeout;

			$qs = 'h='.$server[self::host].'&d='.$this->database[$source].'&q='.rawurlencode($query).'&p='.rawurlencode(serialize($params)).'&u='.(isset($params[1]) && isset($this->distinctrows[$params[1]]) ? rawurlencode(serialize($this->distinctrows[$params[1]])) : serialize(array())).'&k='.md5($this->database[$source].'|'.$server[self::user].'|'.$server[self::pass].'|'.self::version);
			if ($http->post($server[self::http], $qs) === false || strpos($http->contents, 'error') === 0)
			{
				$this->errors[] = '-- HTTP request error ('.$http->status.' :: '.$http->error.' :: '.$qs.' :: '.serialize($http->headers).' :: '.htmlspecialchars($http->contents).') on '.$server[self::http];
				if ($http->status === 0)
				{
					$this->errors[] = '-- Try to limit Database_Checker::rows, or increase Database_Checker::timeout';
				}
				if (self::debug === true)
				{
					$this->infos[] = '-- HTTP request error ('.$http->status.' :: '.$http->error.' :: '.$qs.' :: '.serialize($http->headers).' :: '.htmlspecialchars($http->contents).') on '.$server[self::http];
				}
				return false;
			}
			$content = urldecode($http->contents);
			$content = unserialize($content);
			if ($content === false)
			{
				$this->errors[] = '-- HTTP request error (serialize) on '.$server[self::http].' «'.urldecode($http->contents).'»';
				return false;
			}
			if (isset($content['X-Error']) && isset($content['X-Errors']))
			{
				$this->errors[] = $content['X-Errors'];
				return false;
			}
			return $content;
		}

		$connection = $this->getConnection($source);
		if ($connection === false)
		{
			return false;
		}

		$key = false;
		switch ($query)
		{
			case 'pk':
				{
					$d = $this->getTableData($source, $params[1]);
					$columns = array_keys($d['columns']);
					if (isset($params['skip']))
					{
						sort($columns);
						sort($params['skip']);
						$columns = array_diff($columns, $params['skip']);
					}
					$sql = 'SELECT CONCAT(IFNULL('.implode(',\'NULL\'),\''.self::pkChars.'\',IFNULL(', $d['unique_key']).',\'NULL\')) _PK, CRC32(CONCAT_WS(\',\','.implode(',', $columns).')) C FROM `%s`.`%s` ORDER BY '.implode(', ', $d['unique_key']).' LIMIT %d,%d';
					$key = '_PK';
					break;
				}
			case 'full':
				{
					$d = $this->getTableData($source, $params[1]);
					/* Here we use UNION selects' as if the PK is on multiple columns, a IN() can't be used. The UNION is better than an OR, as MySQL can improve it much better (this includes faster reply from MySQL) */
					$select = 'SELECT *, CONCAT(IFNULL('.implode(',\'NULL\'),\''.self::pkChars.'\',IFNULL(', $d['unique_key']).',\'NULL\')) _PK FROM `%s`.`%s` WHERE '.implode(' /*NULL*/ = %s AND ', $d['unique_key']).' /*NULL*/ = %s UNION ';
					$sql = str_repeat($select, count($params) / (count($d['unique_key']) + 2));
					$sql = substr($sql, 0, -7);
					$key = '_PK';
					break;
				}
			case 'count':
				{
					$sql = 'SELECT COUNT(*) C FROM `'.$this->database[$source].'`.`%s`';
					break;
				}
			case 'tables':
				{
					$sql = 'SHOW TABLES FROM `%s`';
					break;
				}
			case 'table':
				{
					$sql = 'SHOW CREATE TABLE `'.$this->database[$source].'`.`%s`';
					break;
				}
			case 'status':
				{
					$sql = 'SHOW TABLE STATUS FROM `'.$this->database[$source].'` LIKE \'%s\'';
					break;
				}
			case 'checksum':
				{
					$sql = 'CHECKSUM TABLE `'.$this->database[$source].'`.`%s`';
					break;
				}
			case 'broken':
				{
					$sql = 'SELECT f.%s fk FROM `'.$this->database[$source].'`.`%s` f LEFT JOIN `'.$this->database[$source].'`.`%s` m ON (f.%s = m.%s) WHERE f.%s IS NOT NULL AND m.%s IS NULL LIMIT %s,%s';
					break;
				}
			default:
				{
					$this->errors[] = '-- Unknown query: '.str_replace("\n", ' ', $query);
					return false;
				}
		}

		$sql = vsprintf($sql, $params);
		$sql = str_replace('/*NULL*/ = \'NULL\'', 'IS NULL', $sql);
		if (self::debug === true)
		{
			$this->infos[] = '-- Request on '.($source === true ? 'source' : 'destination').': '.$sql;
		}
		if ($this->extension === 'mysql')
		{
			$result = mysql_query($sql, $connection);
		}
		else
		{
			$result = $connection->query($sql);
		}
		if ($result === false)
		{
			if ($this->extension === 'mysql')
			{
				$this->errors[] = '-- Request error: '.str_replace("\n", ' ', $sql).' - '.mysql_error($connection);
			}
			else
			{
				$this->errors[] = '-- Request error: '.str_replace("\n", ' ', $sql).' - '.$connection->error;
			}
			return false;
		}
		$results = array();
		if ($query === 'tables')
		{
			while ($row = ($this->extension === 'mysql' ? mysql_fetch_assoc($result) : $result->fetch_assoc()))
			{
				list(, $t) = each($row);
				$table = $this->query($source, 'table', array($t));
				if (!isset($table[0]['Create Table']))
				{
					$this->errors[] = '-- No create definition for table '.$t;
					return false;
				}
				$create = $table[0]['Create Table'];
				/* Add the list of foreign tables (to be used almost everywhere) */
				$max = preg_match_all('~CONSTRAINT\s+`.+?`\s+FOREIGN\s+KEY\s+.*?\s+REFERENCES\s+`(.+?)`~', $create, $matches);
				$foreignTables = array();
				for ($i = 0; $i < $max; $i++)
				{
					if ($t !== $matches[1][$i])
					{
						$foreignTables[$matches[1][$i]] = true;
					}
				}
				$results[$t] = array($create, $foreignTables);
			}
		}
		else
		{
			while ($row = ($this->extension === 'mysql' ? mysql_fetch_assoc($result) : $result->fetch_assoc()))
			{
				if ($key !== false)
				{
					$results[$row[$key]] = $row;
				}
				else
				{
					$results[] = $row;
				}
			}
		}

		if ($this->extension === 'mysql')
		{
			mysql_free_result($result);
		}
		else
		{
			$result->free();
		}
		return $results;
	}

	/**
	 * Escape some characters to display HTML correctly
	 *
	 * @param string $sql
	 * @param array $data Multiple information about the type of the column
	 * @param boolean $where Whether the escape is for a WHERE or not
	 * @return string
	 */
	private function escape($sql, $data = false, $where = false)
	{
		/* For text fields, we look for non-ASCII chars, then encode the field in Hexadecimal string, with an UNHEX() in SQL */
		if ($data !== false && is_numeric($sql) === false)
		{
			$hex = '';
			/* BLOB and BINARY don't need tests with UTF-8 chars */
			if (strpos($data['type'], 'blob') !== false || strpos($data['type'], 'binary') !== false)
			{
				for ($i = 0, $len = strlen($sql); $i < $len; $i++)
				{
					$hex .= sprintf('%02X', ord($sql{$i}));
				}
				return "UNHEX('".$hex."')";
			}
			for ($i = 0, $len = strlen($sql); $i < $len; $i++)
			{
				$h = ord($sql{$i});
				/* Non-breaking space (&nbsp;) makes copy-paste fail, hence needs to be encoded in Hex */
				if ($h === 160)
				{
					break;
				}

				/** @see http://www.php.net/manual/en/function.ord.php#78032 */
				if ($h >= 0xC2 && $h <= 0xF4)
				{
					break;
				}
			}
			if ($i !== $len)
			{
				for ($i = 0, $len = strlen($sql); $i < $len; $i++)
				{
					$h = ord($sql{$i});
					$hex .= sprintf('%02X', $h);
				}
				return "CONVERT(UNHEX('".$hex."') USING ".$this->charset.")";
			}
		}
		$sql = "'".str_replace(array('\\', '\'', "\r", "\t", "\n"), array('\\\\', '\'\'', '\\r', '\\t', '\\n'), $sql)."'";
		if ($where === true && $data !== false && $data['charset'] !== false)
		{
			$sql = '_'.$data['charset'].' '.$sql;
		}
		return $sql;
	}

	/**
	 * Run a local request, using GET parameters, if the password is okay
	 *
	 * @return string Serialize of the request
	 */
	public static function run($host, $user, $pass)
	{
		if (!isset($_POST['h']) || !isset($_POST['d']) || !isset($_POST['k']) || !isset($_POST['q']) || !isset($_POST['p']) || !isset($_POST['u']))
		{
			header('Status: 400');
			exit('error0');
		}
		if ($_POST['k'] !== md5($_POST['d'].'|'.$user.'|'.$pass.'|'.self::version))
		{
			header('Status: 400');
			exit('error1');
		}
		if ($_POST['h'] !== $host)
		{
			header('Status: 400');
			exit('error2');
		}
		$params = @unserialize($_POST['p']);
		if ($params === false)
		{
			header('Status: 400');
			exit('error3');
		}
		$db = new Database_Checker();
		$db->distinctrows['_from_http_request'] = unserialize($_POST['u']);
		$db->addServer('check', $host, $user, $pass);
		$db->selectDatabase('check', $_POST['d'], true);
		$result = $db->query(true, $_POST['q'], $params);
		if ($result === false)
		{
			$result = array('X-Error' => true, 'X-Errors' => $db->getErrors());
		}
		unset($params);
		$result = urlencode(serialize($result));
		/* Avoid sending the content as Chunked by sending a content-length */
		$acceptEncoding = isset($_SERVER['HTTP_ACCEPT_ENCODING']) ? $_SERVER['HTTP_ACCEPT_ENCODING'] : '';
		if (false && strpos($acceptEncoding, 'gzip') !== false && function_exists('gzcompress'))
		{
			header('Content-Encoding: gzip');
			$crc32 = crc32($result);
			$return = "\x1f\x8b\x08\x00\x00\x00\x00\x00";
			$return .= substr(gzcompress($result), 0, -4);
			$return .= pack('V', $crc32);
			$return .= pack('V', strlen($result));
		}
		else
		{
			/* If ob_gzhandler is used, must send identity */
			header('Content-Encoding: identity');
			$return = &$result;
		}
		header('Content-Length: '.strlen($return));
		echo $return;
		exit;
	}

	/**
	 * Reverse the order of the servers to run tests backwards
	 */
	public function reverse()
	{
		list($this->database[true], $this->database[false]) = array($this->database[false], $this->database[true]);
		list($this->server[true], $this->server[false]) = array($this->server[false], $this->server[true]);
		list($this->uid[true], $this->uid[false]) = array($this->uid[false], $this->uid[true]);
	}

	private function getTableData($source, $table)
	{
		static $tables = array(), $lru = 0;
		$id = $this->database[$source].'.'.$this->server[$source].'.'.$table;
		if (isset($tables[$id]))
		{
			/* Move the data at the end of the array, so that it's ordered correctly for the LRU */
			$data = &$tables[$id];
			unset($tables[$id]);
			$tables[$id] = &$data;
			return $tables[$id];
		}

		$data = $this->getTables($source);
		if (!isset($data[$table]))
		{
			$this->errors[] = '-- Table «'.$this->database[$source].'.'.$table.'» doesn\'t exist on server «'.$this->server[$source].'»';
			return false;
		}
		/* Add a space at the end so that regular expressions below are easier to write */
		$create = $data[$table][0].' ';
		unset($data);

		/* Remove some data if not used since a certain time */
		if ($lru > self::lru)
		{
			/* Remove the first 10 entries, as the array is already correctly LRU-sorted */
			foreach ($tables as $t => $v)
			{
				unset($tables[$t]);
				$lru--;
				if ($lru < self::lru - 10)
				{
					break;
				}
			}
		}
		
		$data = array('auto' => false, 'auto_value' => false, 'primary' => array(), 'unique_key' => array(), 'index' => array(), 'foreign' => array(), 'charset' => '', 'collate' => '');

		/* Columns' list */
		$countMatches = preg_match_all('~`(.+?)` ([a-z]+)(\(.+?\)|)(?:\s+CHARACTER SET (.+?))?(?:\s+COLLATE (.+?))?(?:\s+(.+?))?(?:,|\n)~', $create, $matches);
		for ($i = 0; $i < $countMatches; $i++)
		{
			$tolower = strtolower($matches[6][$i]);
			$default = strpos($tolower, 'default null');
			if ($default !== false)
			{
				$default = 'NULL';
			}
			else
			{
				$default = strpos($tolower, 'default not null');
				if ($default !== false)
				{
					$default = 'NOT NULL';
				}
				else
				{
					$default = strpos($tolower, 'default current_timestamp');
					if ($default !== false)
					{
						$default = 'CURRENT_TIMESTAMP';
					}
					else
					{
						$default = strpos($tolower, 'default');
						if ($default !== false)
						{
							$pos = strpos($matches[6][$i], '\'', $default) + 1;
							$default = substr($matches[6][$i], $pos, strpos($matches[6][$i], '\'', $pos) - $pos);
						}
					}
				}
			}
			$data['columns'][$matches[1][$i]] =
					array(
						'type' => $matches[2][$i],
						'charset' => !empty($matches[4][$i]) ? $matches[4][$i] : false,
						'collate' => !empty($matches[5][$i]) ? $matches[5][$i] : false,
						'info' => trim($matches[3][$i], '()'),
						'unsigned' => strpos($tolower, 'unsigned') !== false,
						'null' => strpos($tolower, 'not null') !== false ? 'not null' : (strpos($tolower, 'null') !== false ? 'null' : false),
						'default' => $default,
						'auto' => strpos($tolower, 'auto_increment') !== false);
		}

		/* Primary keys */
		if (preg_match('~PRIMARY\s+KEY\s+\((.+?)\)~', $create, $matches) > 0)
		{
			$data['primary'] = explode(',', str_replace('`', '', $matches[1]));
			$data['unique_key'] = $data['primary'];
		}
		/* Indexes */
		$countMatches = preg_match_all('~,\s*([A-Z]*)\s*KEY\s+`(.+?)`\s+\((.+)\)~', $create, $matches);
		for ($i = 0; $i < $countMatches; $i++)
		{
			switch ($matches[1][$i])
			{
				case '': $type = 'INDEX';
					break;
				case 'UNIQUE': $type = 'UNIQUE';
					break;
				case 'FULLTEXT': $type = 'FULLTEXT';
					break;
				default:
					$this->errors[] = '-- Index type «'.$matches[1][$i].'» unknown, on table «'.$this->database[$source].'.'.$table.'» on server «'.$this->server[$source].'»';
					return false;
			}
			$data['index'][$matches[2][$i]] =
					array(
						'CONSTRAINT_TYPE' => $type,
						'COLUMNS' => explode(',', str_replace('`', '', $matches[3][$i])));
			if ($data['unique_key'] === array() && $type === 'UNIQUE')
			{
				$data['unique_key'] = $data['index'][$matches[2][$i]]['COLUMNS'];
			}
		}
		/* No unique index, but the user tells us that the table only have distinct rows */
		if ($data['unique_key'] === array())
		{
			if (isset($this->distinctrows['_from_http_request']))
			{
				$distinctRows = $this->distinctrows['_from_http_request'];
			}
			elseif (isset($this->distinctrows[$table]))
			{
				$distinctRows = $this->distinctrows[$table];
			}
			else
			{
				$distinctRows = false;
			}
			if ($distinctRows !== false)
			{
				foreach ($data['columns'] as $k => $v)
				{
					if (!isset($distinctRows[$k]))
					{
						$data['unique_key'][] = $k;
					}
				}
			}
		}
		/* auto_increment */
		if (preg_match('~`(.+?)`.*?auto_increment~', $create, $match) > 0)
		{
			$data['auto'] = $match[1];
		}
		if (preg_match('~AUTO_INCREMENT=(\d+)~', $create, $match) > 0)
		{
			$data['auto_value'] = intval($match[1]);
		}
		/* Table's engine */
		if (preg_match('~ENGINE=(.+?)\s+~', $create, $matches) > 0)
		{
			$data['engine'] = strtolower($matches[1]);
		}
		/* Table's charset */
		if (preg_match('~CHARSET=(.+?)\s+~', $create, $matches) > 0)
		{
			$data['charset'] = strtolower($matches[1]);
		}
		if (preg_match('~COLLATE=(.+?)\s+~', $create, $matches) > 0)
		{
			$data['collate'] = strtolower($matches[1]);
		}
		/* Create table query */
		$data['create'] = preg_replace('~(CREATE\s+TABLE\s+)`(.+?)`~', '\\1`%s`.`%s`', $create);

		/* Foreign keys */
		$countMatches = preg_match_all('~,\s+CONSTRAINT\s+`(.+?)`\s+FOREIGN\s+KEY\s+\(`(.+?)`\)\s+REFERENCES\s+`(.+?)`\s+\(`(.+?)`\)\s*(?:ON|)\s*(?:DELETE|)\s*(CASCADE|SET NULL|NO ACTION|RESTRICT|)\s*(?:ON|)\s*(?:UPDATE|)\s*(CASCADE|SET NULL|NO ACTION|RESTRICT|)~', $create, $matches);
		for ($i = 0; $i < $countMatches; $i++)
		{
			/* Remove foreign keys from here, these will be checked later on */
			$data['create'] = str_replace($matches[0][$i], '', $data['create']);
			$data['foreign'][$matches[1][$i]] =
					array(
						'COLUMNS' => $matches[2][$i],
						'schema' => $this->database[$source],
						'REFERENCED_TABLE_NAME' => $matches[3][$i],
						'REFERENCED_COLUMN_NAME' => $matches[4][$i],
						'DELETE_RULE' => $matches[5][$i],
						'UPDATE_RULE' => $matches[6][$i]);
		}

		$tables[$id] = &$data;
		$lru++;
		return $data;
	}

	/**
	 * Get the SQL code to create the table, with aliases fixed
	 *
	 * @param string $table
	 * @return string
	 */
	private function getTableCreate($table, $refTable = false)
	{
		$data = $this->getTableData(true, $table);
		/* Replace table name with alias if any */
		$destTable = $this->getAlias($table, $refTable);
		$create['create'] = sprintf($data['create'], $this->database[false], $destTable);
		/* Then prepare the code for foreign keys */
		$create['constraints'] = array();
		foreach ($data['foreign'] as $fk)
		{
			$create['constraints'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$this->getFkDefinition($fk, $destTable);
		}
		return $create;
	}

	/**
	 * Get the SQL code to drop a table
	 *
	 * @param string $table
	 * @return string
	 */
	private function getTableDrop($table)
	{
		/* Replace table name with alias if any */
		$data = $this->getTableData(false, $table);
		$drop['drop'] = 'DROP TABLE `'.$this->database[false].'`.`'.$table.'`';
		/* Then prepare the code for foreign keys */
		$drop['constraints'] = array();
		foreach ($data['foreign'] as $k => $fk)
		{
			$drop['constraints'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$table.'` DROP FOREIGN KEY `'.$k.'`';
		}
		return $drop;
	}

	/**
	 * Return table alias in the destination database
	 *
	 * @param string $table
	 * @param string $refTable
	 * @return string
	 */
	private function getAlias($table, $refTable = false)
	{
		$reference = 0;
		$id = $this->uid[true].$this->uid[false];
		if ($refTable !== false && isset($this->aliases[$id]))
		{
			foreach ($this->aliases[$id] as $a)
			{
				foreach ($a as $k => $t)
				{
					if ($t === $refTable)
					{
						$reference = $k;
						break;
					}
				}
			}
		}
		if (isset($this->aliases[$id][$table]))
		{
			if (isset($this->aliases[$id][$table][$reference]))
			{
				return $this->aliases[$id][$table][$reference];
			}
			else
			{
				list(, $t) = each($this->aliases[$id][$table]);
				return $t;
			}
		}
		return $table;
	}

	/**
	 * Get column's definition
	 *
	 * @param string $column
	 * @param array $data
	 * @return string
	 */
	private function getColumnDefinition($column, $data)
	{
		$d = &$data['columns'][$column];
		$colDef = array($d['type'], array());
		if ($d['info'] !== '')
		{
			$colDef[0] .= '('.$d['info'].')';
		}
		if ($d['charset'] !== false)
		{
			$colDef[0] .= ' CHARACTER SET '.$d['charset'];
		}
		if ($d['collate'] !== false)
		{
			$colDef[0] .= ' COLLATE '.$d['collate'];
		}
		if ($d['unsigned'] === true)
		{
			$colDef[0] .= ' UNSIGNED';
		}
		if ($d['null'] === 'null')
		{
			$colDef[0] .= ' NULL';
		}
		elseif ($d['null'] === 'not null')
		{
			$colDef[0] .= ' NOT NULL';
		}
		if ($d['default'] !== false)
		{
			if ($d['default'] === 'NULL' || $d['default'] === 'NOT NULL')
			{
				$colDef[0] .= ' DEFAULT '.$d['default'];
			}
			else
			{
				$colDef[0] .= ' DEFAULT '.$this->escape($d['default']);
			}
		}
		if ($d['auto'] === true)
		{
			$colDef[0] .= ' auto_increment';
		}
		/* Add indexes on the new column */
		foreach ($data['index'] as $name => $idx)
		{
			if (in_array($column, $idx['COLUMNS']))
			{
				$colDef[1][] = $idx['CONSTRAINT_TYPE'].' `'.$name.'` ('.str_replace(array('(', ')`'), array('`(', ')'), '`'.implode('`,`', $idx['COLUMNS']).'`').')';
			}
		}
		/* Then add foreign keys */
		foreach ($data['foreign'] as $name => $fk)
		{
			if ($fk['COLUMNS'] === $column)
			{
				$colDef[1][] = $this->getFkDefinition($fk, false);
			}
		}
		return $colDef;
	}

	/**
	 * Get foreign key's definition
	 *
	 * @param array $data
	 * @param string $refTable
	 * @return string
	 */
	private function getFkDefinition($data, $refTable)
	{
		$data['REFERENCED_TABLE_NAME'] = $this->getAlias($data['REFERENCED_TABLE_NAME'], $refTable);
		$def = 'FOREIGN KEY (`'.$data['COLUMNS'].'`) REFERENCES `'.$this->database[false].'`.`'.$data['REFERENCED_TABLE_NAME'].'` (`'.$data['REFERENCED_COLUMN_NAME'].'`)';
		if ($data['DELETE_RULE'] !== '')
		{
			$def .= ' ON DELETE '.$data['DELETE_RULE'];
		}
		if ($data['UPDATE_RULE'] !== '')
		{
			$def .= ' ON UPDATE '.$data['UPDATE_RULE'];
		}
		return $def;
	}

	/**
	 * Table definitions' differencies
	 *
	 * @param string $table Source table
	 * @return string
	 */
	public function getTableColumnsDiff($table)
	{
		$return = array('drop_idx' => array(), 'drop' => array(), 'add' => array(), 'change' => array(), 'add_idx' => array(), 'change_idx' => array(), 'order' => array());
		$id = $this->uid[true].$this->uid[false];
		if (isset($this->aliases[$id][$table]))
		{
			$tables = $this->aliases[$id][$table];
		}
		else
		{
			$tables = array($table);
		}
		$sourceData = $this->getTableData(true, $table);
		if ($sourceData === false)
		{
			return false;
		}
		foreach ($tables as $destTable)
		{
			$destData = $this->getTableData(false, $destTable);
			if ($destData === false)
			{
				return false;
			}
			$sourceKeys = array_keys($sourceData['columns']);
			$destKeys = array_keys($destData['columns']);

			/* Check engine */
			if ($sourceData['engine'] !== $destData['engine'])
			{
				$this->infos[] = '-- Wrong engine';
				$return['change'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ENGINE = '.$sourceData['engine'];
			}

			/* Check character set and collate */
			if ($sourceData['charset'].'|'.$sourceData['collate'] !== $destData['charset'].'|'.$destData['collate'])
			{
				$this->infos[] = '-- Wrong charset/collate';
				$return['change'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DEFAULT CHARACTER SET '.$sourceData['charset'].(!empty($sourceData['collate']) ? ' COLLATE '.$sourceData['collate'] : '');
			}

			/* Wrong order of columns: should be the same columns at least */
			if ($sourceKeys !== $destKeys)
			{
				$sourceCheck = $sourceKeys;
				sort($sourceCheck);
				$destCheck = $destKeys;
				sort($destCheck);
				if ($sourceCheck === $destCheck)
				{
					/* Keys are the same after sorting, so the order is wrong */
					for ($i = 0, $max = count($sourceKeys); $i < $max; $i++)
					{
						if ($sourceKeys[$i] !== $destKeys[$i])
						{
							break;
						}
					}
					if ($i === 0)
					{
						$lastKey = 'FIRST';
					}
					else
					{
						$lastKey = 'AFTER '.$sourceKeys[$i - 1];
					}
					$this->infos[] = '-- Wrong columns order';
					/* Search for any foreign key */
					foreach ($destData['foreign'] as $k => $fk)
					{
						if ($fk['COLUMNS'] === $sourceKeys[$i])
						{
							$return['order'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP FOREIGN KEY '.$k;
						}
					}
					$colDef = $this->getColumnDefinition($sourceKeys[$i], $sourceData);
					$return['order'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` CHANGE '.$sourceKeys[$i].' '.$sourceKeys[$i].'_XYZ '.$colDef[0];
					$return['order'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$sourceKeys[$i].' '.$colDef[0].' '.$lastKey;
					$return['order'][] = 'UPDATE `'.$this->database[false].'`.`'.$destTable.'` SET '.$sourceKeys[$i].' = '.$sourceKeys[$i].'_XYZ';
					$return['order'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP '.$sourceKeys[$i].'_XYZ';
					/* Add indexes and foreign keys */
					foreach ($colDef[1] as $idx)
					{
						$this->infos[] = '-- Add index or foreign key';
						$return['order'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$idx;
					}
				}
			}

			/* Missing column */
			$diffCols = array_diff($sourceKeys, $destKeys);
			foreach ($diffCols as $column)
			{
				/* Searching for previous column */
				$lastKey = '';
				for ($i = 0, $max = count($sourceKeys); $i < $max; $i++)
				{
					if ($sourceKeys[$i] === $column)
					{
						break;
					}
					$lastKey = $sourceKeys[$i];
				}
				if ($i === $max || $lastKey === '')
				{
					$lastKey = ' FIRST';
				}
				else
				{
					$lastKey = ' AFTER '.$lastKey;
				}
				$this->infos[] = '-- Missing column '.$column;
				$colDef = $this->getColumnDefinition($column, $sourceData);
				$sql = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$column.' '.$colDef[0];
				if ($sourceData['auto'] === $column)
				{
					$sql .= ' AUTO_INCREMENT'.$lastKey.', ADD PRIMARY KEY ('.$column.')';
				}
				else
				{
					$sql .= $lastKey;
				}
				$return['add'][] = $sql;
				foreach ($colDef[1] as $idx)
				{
					$this->infos[] = '-- Add index or foreign key';
					$return['add'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$idx;
				}
			}

			/* Extra column */
			$diffArray = array_diff($destKeys, $sourceKeys);
			foreach ($diffArray as $column)
			{
				$this->infos[] = '-- Extra column '.$column;
				$return['drop'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP '.$column;
			}

			/* Checking columns' types */
			$commonCols = array_intersect($sourceKeys, $destKeys);
			foreach ($commonCols as $c)
			{
				if ($sourceData['columns'][$c] !== $destData['columns'][$c])
				{
					$colDef = $this->getColumnDefinition($c, $sourceData);
					$this->infos[] = '-- Wrong colum definition '.$c;
					$return['change'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` CHANGE '.$c.' '.$c.' '.$colDef[0];
				}
			}

			/* Primary keys */
			if ($sourceData['primary'] !== $destData['primary'])
			{
				if ($sourceData['primary'] === array())
				{
					$this->infos[] = '-- Extra primary';
					$return['drop'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP PRIMARY KEY';
				}
				elseif ($destData['primary'] === array())
				{
					$this->infos[] = '-- Missing primary';
					$return['add'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD PRIMARY KEY (`'.implode('`,`', $sourceData['primary']).'`)';
				}
				else
				{
					$this->infos[] = '-- Wrong primary key';
					$return['change'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP PRIMARY KEY, ADD PRIMARY KEY (`'.implode('`,`', $sourceData['primary']).'`)';
				}
				if (count($sourceData['primary']) !== 0)
				{
					foreach ($destData['index'] as $idx => $v)
					{
						/* If a change is made on the primary key, then if an index exists on the column of the PK, it's automatically removed, so don't check for indexes after this primary change */
						if (count($v['COLUMNS']) === 1 && $v['COLUMNS'][0] === $sourceData['primary'][0])
						{
							unset($destData['index'][$idx]);
						}
					}
				}
			}

			/* Foreign keys */
			$sourceFk = array();
			$destFk = array();
			foreach ($sourceData['foreign'] as $k => $key)
			{
				if (isset($sourceFk[$key['COLUMNS']]))
				{
					$this->infos[] = '-- Duplicate foreign key on '.$key['COLUMNS'].' ignored, you may delete it using ALTER TABLE '.$table.' DROP FOREIGN KEY `'.$k.'`';
					continue;
				}
				$sourceFk[$key['COLUMNS']] = $key;
				$sourceFk[$key['COLUMNS']]['name'] = $k;
			}
			foreach ($destData['foreign'] as $k => $key)
			{
				if (isset($destFk[$key['COLUMNS']]))
				{
					$this->infos[] = '-- Duplicate foreign key on '.$key['COLUMNS'];
					$return['drop_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP FOREIGN KEY `'.$k.'`';
					continue;
				}
				$destFk[$key['COLUMNS']] = $key;
				$destFk[$key['COLUMNS']]['name'] = $k;
			}
			$indexArray = $sourceFk + $destFk;
			foreach ($indexArray as $idx => $v)
			{
				if (!isset($sourceFk[$idx]))
				{
					$this->infos[] = '-- Extra foreign key on '.$this->database[false].'.'.$destTable.': '.$idx;
					$return['drop_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP FOREIGN KEY `'.$v['name'].'`';
				}
				elseif (!isset($destFk[$idx]))
				{
					/* If no missing column exists in the index, then add it (else the index is created just before as we add the key) */
					if (!in_array($v['REFERENCED_COLUMN_NAME'], $diffCols))
					{
						$this->infos[] = '-- Missing foreign key on '.$this->database[false].'.'.$destTable.': '.$idx;
						$return['add_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$this->getFkDefinition($v, $destTable);
					}
				}
				else
				{
					$sourceCheck = $sourceFk[$idx];
					$sourceCheck['REFERENCED_TABLE_NAME'] = $this->getAlias($sourceCheck['REFERENCED_TABLE_NAME'], $destTable);
					unset($sourceCheck['name']);
					unset($sourceCheck['schema']);
					$destCheck = $destFk[$idx];
					unset($destCheck['name']);
					unset($destCheck['schema']);
					if ($sourceCheck !== $destCheck)
					{
						$this->infos[] = '-- Wrong foreign key: '.implode('|', $sourceCheck).' vs '.implode('|', $destCheck);
						$return['change_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP FOREIGN KEY `'.$destFk[$idx]['name'].'`, ADD '.$this->getFkDefinition($v, $destTable);
					}
				}
			}

			/* Indexes */
			$indexArray = $sourceData['index'] + $destData['index'];
			foreach ($indexArray as $idx => $v)
			{
				if (!isset($sourceData['index'][$idx]))
				{
					$this->infos[] = '-- Extra index '.$idx;
					$return['drop_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP INDEX `'.$idx.'`';
				}
				elseif (!isset($destData['index'][$idx]))
				{
					/* If no missing column exists in the index, then add it (else the index is created just before as we add the key) */
					if (array_intersect($sourceData['index'][$idx]['COLUMNS'], $diffCols) === array())
					{
						$this->infos[] = '-- Missing index '.$idx;
						$return['add_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` ADD '.$sourceData['index'][$idx]['CONSTRAINT_TYPE'].' `'.$idx.'` ('.str_replace(array('(', ')`'), array('`(', ')'), '`'.implode('`,`', $sourceData['index'][$idx]['COLUMNS']).'`').')';
					}
				}
				elseif ($sourceData['index'][$idx] !== $destData['index'][$idx])
				{
					$this->infos[] = '-- Wrong index '.$idx;
					$return['change_idx'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` DROP INDEX `'.$idx.'`, ADD '.$sourceData['index'][$idx]['CONSTRAINT_TYPE'].' `'.$idx.'` ('.str_replace(array('(', ')`'), array('`(', ')'), '`'.implode('`,`', $sourceData['index'][$idx]['COLUMNS']).'`').')';
				}
			}
		}
		return $this->format($return);
	}

	/**
	 * Returns differences between two databases (missing tables)
	 *
	 * @param string $sourceDb		Source database
	 * @param string $destDb		Destination database
	 */
	public function getDatabaseDiff()
	{
		$return = array('drop_idx' => array(), 'drop' => array(), 'create' => array(), 'add_idx' => array());
		$sourceTables = $this->getTables(true, true);
		$destTables = $this->getTables(false, true);
		if ($sourceTables === false || $destTables === false)
		{
			return false;
		}
		$reverseTables = array();
		$id = $this->uid[true].$this->uid[false];
		foreach ($sourceTables as $k => $t)
		{
			if (isset($this->aliases[$id][$t]))
			{
				foreach ($this->aliases[$id][$t] as $t1)
				{
					if (!is_null($t1))
					{
						$sourceTables[] = $t1;
						$reverseTables[$t1] = $t;
					}
				}
				unset($sourceTables[$k]);
			}
			else
			{
				$reverseTables[$t] = $t;
			}
		}
		/* Must reverse arrays, as latest table has the more foreign keys with other tables (see getTables for the order of tables) */
		$sourceTables = array_reverse($sourceTables);
		$destTables = array_reverse($destTables);
		$diff = array_diff($sourceTables, $destTables);
		foreach ($diff as $d)
		{
			$create = $this->getTableCreate($reverseTables[$d], $d);
			if (count($create['constraints']) !== 0)
			{
				$this->infos[] = '-- Missing constraints for table '.$d;
				$return['add_idx'] = array_merge($return['add_idx'], $create['constraints']);
			}
			$this->infos[] = '-- Missing table '.$d;
			$return['create'][] = $create['create'];
		}
		$diff = array_diff($destTables, $sourceTables);
		foreach ($diff as $d)
		{
			$drop = $this->getTableDrop($d);
			if (count($drop['constraints']) !== 0)
			{
				$this->infos[] = '-- Delete constraints for table '.$d.' before deleting the table';
				$return['drop_idx'] = array_merge($return['drop_idx'], $drop['constraints']);
			}
			$this->infos[] = '-- Extra table '.$d;
			$return['drop'][] = $drop['drop'];
		}
		return $this->format($return);
	}

	/**
	 * Tables' list
	 *
	 * @param boolean $fromSource
	 * @param boolean $onlyTables Return only tables without create info
	 * @return array
	 */
	public function getTables($fromSource = true, $onlyTables = false)
	{
		static $tables = array();
		$id = $this->database[$fromSource].'.'.$this->server[$fromSource];
		if (!isset($tables[$id]))
		{
			$data = $this->query($fromSource, 'tables', array($this->database[$fromSource]));
			if ($data === false)
			{
				return false;
			}
			$count = count($data);
			$tables[$id] = array();
			if ($count !== 0)
			{
				/* Now that we have tables, we must compute the order, as foreign keys may impact */
				$tableSorted = array_combine(array_keys($data), array_fill(0, $count, 1));
				$tableLinks = array_combine(array_keys($data), array_fill(0, $count, array()));
				foreach ($data as $table => $create)
				{
					foreach ($create[1] as $t => $v)
					{
						/* Search for tables with mutual links (and ignore these links) */
						if (isset($data[$t][1][$table]))
						{
							unset($data[$t][1][$table]);
							unset($data[$table][1][$t]);
						}
						else
						{
							/* Else just add a link to the foreign table */
							$tableLinks[$table][] = &$tableSorted[$t];
						}
					}
				}
				$found = 1;
				while ($found !== 0)
				{
					$found = 0;
					foreach ($tableLinks as $table => $d)
					{
						/* If there's no more table linked to this one, we can add it to the list */
						if (array_sum($d) === 0)
						{
							$tableSorted[$table] = 0;
							$tables[$id][$table] = &$data[$table];
							unset($tableLinks[$table]);
							$found++;
						}
					}
				}
				/* Some links are bidirectional, so these can't be removed with the above computing */
				foreach ($tableLinks as $table => $d)
				{
					$tables[$id][$table] = &$data[$table];
				}
				unset($tableLinks);
				unset($tableSorted);
			}
		}
		if ($onlyTables === true)
		{
			return array_keys($tables[$id]);
		}
		return $tables[$id];
	}

	/**
	 * Get data checksum for a table
	 *
	 * @param boolean $fromSource
	 * @param string $table
	 * @return string
	 */
	private function getTableChecksum($fromSource, $table)
	{
		static $checksums = array();
		if (self::checksum === false)
		{
			return false;
		}
		$id = $this->database[$fromSource].'.'.$this->server[$fromSource].'.'.$table;
		if (!isset($checksums[$id]))
		{
			$checksum = $this->query($fromSource, 'checksum', array($table));
			if ($checksum === false)
			{
				return false;
			}
			$checksums[$id] = $checksum[0]['Checksum'];
		}
		return $checksums[$id];
	}

	/**
	 * Table contents' differencies
	 *
	 * @param string	$sourceTable Source table
	 * @param array		$reverse Results in the other way
	 * @return array
	 */
	public function getTableDiff($sourceTable, &$reverse)
	{
		/* At least a little bit of memory must be available */
		if ($this->memory < 0.2)
		{
			$this->errors[] = '-- Not enough memory to run diff on next table: '.$sourceTable;
			return false;
		}
		$return = $reverse = array('delete' => array(), 'update' => array(), 'insert' => array(), 'auto' => array());
		$id = $this->uid[true].$this->uid[false];
		if (isset($this->aliases[$id][$sourceTable]))
		{
			$destTables = $this->aliases[$id][$sourceTable];
		}
		else
		{
			$destTables = array($sourceTable);
		}
		$sourceData = $this->getTableData(true, $sourceTable);
		if ($sourceData === false)
		{
			return false;
		}
		$sourcePk = &$sourceData['unique_key'];
		$sourceColumns = array_keys($sourceData['columns']);
		/* Listing key (must be a primary key or unique key) */
		if (count($sourcePk) === 0)
		{
			$this->errors[] = '-- Can\'t find any primary key on '.$sourceTable.', you may use '.__CLASS__.'::setDistinctRows()';
			return false;
		}

		/* Looking for the checksum */
		$sourceChecksum = $this->getTableChecksum(true, $sourceTable);
		if ($sourceChecksum === false)
		{
			$this->errors[] = '-- Checksum can\'t be computed on table '.$sourceTable;
			return false;
		}

		foreach ($destTables as $destTable)
		{
			$destData = $this->getTableData(false, $destTable);
			if ($destData === false)
			{
				return false;
			}
			$destPk = &$destData['unique_key'];
			$destColumns = array_keys($destData['columns']);

			/* Looking for the checksum */
			$destChecksum = $this->getTableChecksum(false, $destTable);
			if ($destChecksum === false)
			{
				$this->errors[] = '-- Checksum can\'t be computed on table '.$destTable;
				return false;
			}

			$rowNumber = false;
			/* If the option to cache the row number exists, use it, if the file isn't too old (1 hour) */
			if ($this->rowNumberCacheDir !== '')
			{
				/* Yes, there's a risk that MD5 could be the same for 2 different strings, but that's no big deal here */
				$rowNumberFile = $this->rowNumberCacheDir.md5($id.'.'.$sourceTable.'.'.$destTable).'.database_checker.tmp';
				if (file_exists($rowNumberFile) && filemtime($rowNumberFile) > $_SERVER['REQUEST_TIME'] - 3600)
				{
					$rowNumber = (int) file_get_contents($rowNumberFile);
					if ($rowNumber <= 0)
					{
						$rowNumber = -1;
					}
				}
				else
				{
					$rowNumber = -1;
				}
			}
			if ($sourceChecksum === $destChecksum)
			{
				/* Remove the cache on this table, as tables are the same */
				if ($rowNumber !== false && file_exists($rowNumberFile))
				{
					unlink($rowNumberFile);
				}
				if (self::debug === true)
				{
					$this->infos[] = '-- Same checksum';
				}
				continue;
			}
			if (self::debug === true)
			{
				$this->infos[] = '-- Checksum mismatch: '.$sourceChecksum.' vs '.$destChecksum;
			}

			/* Check auto_increment */
			if ($sourceData['auto_value'] !== false && $sourceData['auto_value'] !== $destData['auto_value'])
			{
				$this->infos[] = '-- Wrong autoincrement';
				if (self::autoincrement === true)
				{
					$return['auto'][] = 'ALTER TABLE `'.$this->database[false].'`.`'.$destTable.'` AUTO_INCREMENT = '.$sourceData['auto_value'];
					$reverse['auto'][] = 'ALTER TABLE `'.$this->database[true].'`.`'.$sourceTable.'` AUTO_INCREMENT = '.$destData['auto_value'];
				}
			}

			if ($sourceColumns !== $destColumns)
			{
				$this->errors[] = '-- Tables\' columns are different: '.implode(', ', $sourceColumns).' vs '.implode(', ', $destColumns);
				return false;
			}
			if ($sourcePk !== $destPk)
			{
				$this->errors[] = '-- Tables\' primary keys are different: '.implode(', ', $sourcePk).' vs '.implode(', ', $destPk);
				return false;
			}

			/* We're looking for the total rows in the table. This way, we can start from the end, which is most probably the location of changed rows */
			if (($sourceRows = $this->query(true, 'count', array($sourceTable))) === false || ($destRows = $this->query(false, 'count', array($destTable))) === false)
			{
				$this->errors[] = '-- Can\'t find total count of table : '.$sourceTable.' or '.$destTable;
				return false;
			}

			$sourceRows = (int) floor($sourceRows[0]['C'] / self::keys);
			$destRows = (int) floor($destRows[0]['C'] / self::keys);
			/* If the row's number seems good, use it */
			if ($rowNumber !== false && $rowNumber >= 0 && $rowNumber <= max($sourceRows, $destRows))
			{
				$rows = $rowNumber;
				if (self::debug === true)
				{
					$this->infos[] = '-- Found cached row\'s number: '.$rowNumber;
				}
			}
			elseif ($sourceRows === $destRows)
			{
				/* Same number of rows ? Just use it :-) */
				$rows = $sourceRows;
			}
			else
			{
				/**
				 * We'll first try to add all missing rows. This means start with the first row not available : min($sourceRows, $destRows). Add the number of differencies asked for (self::diffs), because we'll run at least this number of rows after the first empty one
				 * Finally, we take the greatest value between the number of rows in both table, and use it to limit our previous value.
				 */
				$rows = (int) min(max($sourceRows, $destRows), min($sourceRows, $destRows) + self::diffs);
			}

			$sParams = array($this->database[true], $sourceTable, 0, self::keys);
			$dParams = array($this->database[false], $destTable, 0, self::keys);
			if (isset($this->skip[$sourceTable]))
			{
				$sParams['skip'] = $dParams['skip'] = array_keys($this->skip[$sourceTable]);
			}

			/**
			 * The process is to order tables by PK, get self::keys rows each time, and find what lines are the same.
			 *
			 * This is far from ideal, but as many people won't have an easy key (such as and integer key), this is the faster way to find differencies when tables are about the same, but it will be longer if not (as we'll check rows against all rows left in the queue waiting for the same row on the source/dest table), but there's no easy way to do this as we don't have a memory unlimited process.
			 *
			 * Improvements are _really_ welcome here :-)
			 */
			$foundDiffs = 0;
			while ($rows >= 0)
			{
				/* We've seen more than self::diffs blocks of rows after the last change, let's break out of the loop. This test must be here because there's some continue() in the loop */
				if ($foundDiffs !== 0 && $foundDiffs++ >= self::diffs)
				{
					break;
				}

				$sParams[2] = $dParams[2] = $rows * self::keys;
				if (($s = $this->query(true, 'pk', $sParams)) === false || ($d = $this->query(false, 'pk', $dParams)) === false)
				{
					return false;
				}
				$rows--;
				/* Both requests are the same, ignore these and go on to the next one */
				if ($s === $d)
				{
					continue;
				}
				foreach ($s as $k => $v)
				{
					/* Same key exists */
					if (isset($d[$k]))
					{
						if ($v === $d[$k])
						{
							/* Value is the same, remove it */
							unset($s[$k]);
							unset($d[$k]);
						}
					}
				}
				$diffs = array_merge(array_keys($s), array_keys($d));

				$max = count($diffs);

				$found = false;
				$inserts = $reverseInserts = array();
				for ($i = 0; $i < $max; $i += self::rows)
				{
					/* Now it's time to select content differences. Code is about the same, but that's to minimize the data queried on MySQL on the first call */
					$keys = array_slice($diffs, $i, self::rows);
					$db = $this->database[true];
					$tb = $sourceTable;
					$params = array();
					foreach ($keys as $k)
					{
						$k = explode(self::pkChars, $k);
						$params[] = &$db;
						$params[] = &$tb;
						foreach ($k as $u => $v)
						{
							$params[] = $this->escape($v, $sourceData['columns'][$sourcePk[$u]], true);
						}
					}
					$sourceArray = $this->query(true, 'full', $params);
					$db = $this->database[false];
					$tb = $destTable;
					$destArray = $this->query(false, 'full', $params);
					if ($sourceArray === false || $destArray === false)
					{
						return false;
					}
					$listArray = $sourceArray + $destArray;
					$keys = array();
					$characters = 0;
					foreach ($listArray as $k => $value)
					{
						$s = &$sourceArray[$k];
						$d = &$destArray[$k];

						$where = '';
						foreach ($sourcePk as $k1)
						{
							if (isset($s))
							{
								$where .= $k1.(is_null($s[$k1]) ? ' IS NULL' : ' = '.$this->escape($s[$k1], $sourceData['columns'][$k1])).' AND ';
							}
							else
							{
								$where .= $k1.(is_null($d[$k1]) ? ' IS NULL' : ' = '.$this->escape($d[$k1], $sourceData['columns'][$k1])).' AND ';
							}
						}
						$where = substr($where, 0, -5);
						/* Looking for updates */
						if (isset($d) && isset($s))
						{
							/* Same values, ignore this */
							if ($s === $d)
							{
								continue;
							}

							unset($s['_PK']);
							$updateDest = $updateSource = '';
							foreach ($s as $k1 => &$v)
							{
								if ($v !== $d[$k1])
								{
									$updateDest .= $k1.' = '.(is_null($v) ? 'NULL, ' : htmlspecialchars($this->escape($v, $sourceData['columns'][$k1])).', ');
									$updateSource .= $k1.' = '.(is_null($d[$k1]) ? 'NULL, ' : htmlspecialchars($this->escape($d[$k1], $sourceData['columns'][$k1])).', ');
								}
							}
							$characters += strlen($updateDest) * 2; // size of SET part
							$return['update'][] = 'UPDATE `'.$this->database[false].'`.`'.$destTable.'` SET '.substr($updateDest, 0, -2).' WHERE '.$where;
							$reverse['update'][] = 'UPDATE `'.$this->database[true].'`.`'.$sourceTable.'` SET '.substr($updateSource, 0, -2).' WHERE '.$where;
						}
						/* Then, it must be an INSERT */
						elseif (isset($s))
						{
							$insert = '';
							unset($s['_PK']);
							foreach ($s as $k1 => &$v)
							{
								$insert .= ( is_null($v) ? 'NULL, ' : htmlspecialchars($this->escape($v, $sourceData['columns'][$k1])).', ');
							}
							$characters += strlen($insert);
							$inserts[] = substr($insert, 0, -2);
							$reverse['delete'][] = 'DELETE FROM `'.$this->database[true].'`.`'.$sourceTable.'` WHERE '.$where;
						}
						/* Sure, it's a DELETE */
						else
						{
							$return['delete'][] = 'DELETE FROM `'.$this->database[false].'`.`'.$destTable.'` WHERE '.$where;
							$insert = '';
							unset($d['_PK']);
							foreach ($d as $k1 => &$v)
							{
								$insert .= ( is_null($v) ? 'NULL, ' : htmlspecialchars($this->escape($v, $sourceData['columns'][$k1])).', ');
							}
							$characters += strlen($insert);
							$reverseInserts[] = substr($insert, 0, -2);
						}
						$characters += strlen($where) * 2 + 100; // size of main SQL string + where clause
						$found = true;
					}
					unset($s);
					unset($d);
					unset($sourceArray);
					unset($destArray);
					/**
					 * Compute RAM used to store all this text:
					 * - rough number of characters
					 * - test made for an array of 10000 entries of 1000 chars gives 12692080 bytes, hence a secured value of 1.5 byte for 1 char, so 0.0000015 Mb for 1 char
					 * - set foundDiffs so that the next loop will break
					 */
					$this->memory -= $characters * 0.0000015;
					if ($this->memory < 0)
					{
						$foundDiffs = self::diffs;
						break;
					}
				}
				if ($found === true)
				{
					/* Group INSERTS so that it's faster while inserting in the database */
					while (count($inserts) !== 0)
					{
						$splice = array_splice($inserts, 0, self::inserts);
						$return['insert'][] = 'INSERT INTO `'.$this->database[false].'`.`'.$destTable.'` VALUES ('.implode('),(', $splice).')';
					}
					unset($inserts);
					while (count($reverseInserts) !== 0)
					{
						$splice = array_splice($reverseInserts, 0, self::inserts);
						$reverse['insert'][] = 'INSERT INTO `'.$this->database[true].'`.`'.$sourceTable.'` VALUES ('.implode('),(', $splice).')';
					}
					unset($reverseInserts);
					/* Found at least one diff, start the diff counter (see self::diffs) */
					if ($foundDiffs === 0)
					{
						$foundDiffs = 1;
					}
					$found = false;
				}
			}
			/* Save the row's number, if asked for */
			if ($rowNumber !== false)
			{
				if ($rows < 0 || $sourceRows !== $destRows)
				{
					if (file_exists($rowNumberFile))
					{
						unlink($rowNumberFile);
					}
				}
				else
				{
					file_put_contents($rowNumberFile, ($rows + 1)."\n".$sourceTable."\n");
				}
			}
		}
		$reverse = $this->format($reverse);
		return $this->format($return);
	}

	/**
	 * Search for broken foreign keys
	 *
	 * @param string $table
	 * @return array
	 */
	public function findBrokenKeys($table)
	{
		$return = array('delete' => array());
		$data = $this->getTableData(true, $table);
		if ($data === false)
		{
			return false;
		}
		foreach ($data['foreign'] as $foreign)
		{
			$missing = array();
			for ($i = 0;; $i += self::keys)
			{
				/* Run a query with LEFT JOIN, WHERE join_key is NULL */
				$keys = $this->query(true, 'broken', array($foreign['COLUMNS'], $table, $foreign['REFERENCED_TABLE_NAME'], $foreign['COLUMNS'], $foreign['REFERENCED_COLUMN_NAME'], $foreign['COLUMNS'], $foreign['REFERENCED_COLUMN_NAME'], $i, $i + self::keys));
				foreach ($keys as $k)
				{
					$missing[$k['fk']] = true;
				}
				if (count($keys) !== self::keys)
				{
					break;
				}
			}
			if (!empty($missing))
			{
				if ($data['columns'][$foreign['COLUMNS']]['null'] === 'null')
				{
					/* If the column is nullable, then set it to NULL */
					foreach ($missing as $k => $v)
					{
						$return['delete'][] = 'UPDATE `'.$this->database[true].'`.`'.$table.'` SET '.$foreign['COLUMNS'].' = NULL WHERE '.$foreign['COLUMNS'].' = '.htmlspecialchars($this->escape($k));
					}
				}
				else
				{
					/* Else delete the current entry */
					foreach ($missing as $k => $v)
					{
						$return['delete'][] = 'DELETE FROM `'.$this->database[true].'`.`'.$table.'` WHERE '.$foreign['COLUMNS'].' '.(is_null($k) ? 'IS NULL' : '= '.htmlspecialchars($this->escape($k)));
					}
				}
			}
		}

		return $this->format($return);
	}

	/**
	 * Returns formatted code
	 *
	 * @param array $results Results to format
	 * @return array
	 */
	private function format($results)
	{
		if ($this->format === 'array')
		{
			return $results;
		}
		foreach ($results as $k => &$value)
		{
			if (count($value) === 0)
			{
				$value = '';
			}
			elseif ($this->format === 'sql')
			{
				$value = implode(";\n", $value).";\n";
			}
			else
			{
				$value = str_replace(' ', '&nbsp;', implode(";<br/>", $value));
			}
		}
		return $results;
	}

	/**
	 * Returns some information about checks
	 */
	public function getInfos()
	{
		$result = $this->format(array($this->infos));
		return $result[0];
	}

	/**
	 * Returns errors about checks
	 */
	public function getErrors()
	{
		$result = $this->format(array($this->errors));
		return $result[0];
	}

	public function __destruct()
	{
		foreach ($this->connections as $conn)
		{
			if ($this->extension === 'mysql')
			{
				mysql_close($conn);
			}
			else
			{
				$conn->close();
			}
		}
	}

}
?>